#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import numpy as np
from sklearn.datasets import make_classification
#Internal libraries
from aucmedi.sampling import sampling_split, sampling_kfold

#-----------------------------------------------------#
#                  Unittest: Sampling                 #
#-----------------------------------------------------#
class SamplingTEST(unittest.TestCase):
    # Create random classification data
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Generate data
        self.x, y = make_classification(n_samples=1000, n_features=4,
                                        random_state=0, shuffle=False)
        # One-hot encode labels
        self.y = np.zeros((y.size, 2))
        self.y[np.arange(y.size), y] = 1

    #-------------------------------------------------#
    #          Sampling via Percentage Split          #
    #-------------------------------------------------#
    # Check percentage split ratio exception handling
    def test_PercentageSplit_samplingCheck(self):
        self.assertRaises(ValueError, sampling_split, self.x, self.y,
                          sampling=[0.8, 0.05, 0.1])

    # Check random sampling via percentage split
    def test_PercentageSplit_random(self):
        subsets = sampling_split(self.x, self.y,
                                 sampling=[0.4, 0.3, 0.05, 0.25],
                                 iterative=False, stratified=False)
        self.assertTrue(len(subsets) == 4)
        self.assertTrue(len(subsets[0]) == 2)
        self.assertTrue(len(subsets[1]) == 2)
        self.assertTrue(len(subsets[2]) == 2)
        self.assertTrue(len(subsets[3]) == 2)
        self.assertTrue(subsets[0][0].shape[0] > 395 and \
                        subsets[0][0].shape[0] < 405)
        self.assertTrue(subsets[0][1].shape[0] > 395 and \
                        subsets[0][1].shape[0] < 405)
        self.assertTrue(subsets[1][0].shape[0] > 295 and \
                        subsets[1][0].shape[0] < 305)
        self.assertTrue(subsets[2][0].shape[0] > 45 and \
                        subsets[2][0].shape[0] < 55)
        self.assertTrue(subsets[3][0].shape[0] > 245 and \
                        subsets[3][0].shape[0] < 255)

    # Check stratified random sampling via percentage split
    def test_PercentageSplit_stratified(self):
      subsets = sampling_split(self.x, self.y,
                               sampling=[0.4, 0.3, 0.05, 0.25],
                               iterative=False, stratified=True)
      self.assertTrue(len(subsets) == 4)
      self.assertTrue(len(subsets[0]) == 2)
      self.assertTrue(len(subsets[1]) == 2)
      self.assertTrue(len(subsets[2]) == 2)
      self.assertTrue(len(subsets[3]) == 2)
      self.assertTrue(subsets[0][0].shape[0] > 395 and \
                      subsets[0][0].shape[0] < 405)
      self.assertTrue(subsets[0][1].shape[0] > 395 and \
                      subsets[0][1].shape[0] < 405)
      self.assertTrue(subsets[1][0].shape[0] > 295 and \
                      subsets[1][0].shape[0] < 305)
      self.assertTrue(subsets[2][0].shape[0] > 45 and \
                      subsets[2][0].shape[0] < 55)
      self.assertTrue(subsets[3][0].shape[0] > 245 and \
                      subsets[3][0].shape[0] < 255)

    # Check stratified iterative sampling via percentage split
    def test_PercentageSplit_iterative(self):
        subsets = sampling_split(self.x, self.y,
                                 sampling=[0.4, 0.3, 0.05, 0.25],
                                 iterative=True, stratified=True)
        self.assertTrue(len(subsets) == 4)
        self.assertTrue(len(subsets[0]) == 2)
        self.assertTrue(len(subsets[1]) == 2)
        self.assertTrue(len(subsets[2]) == 2)
        self.assertTrue(len(subsets[3]) == 2)
        self.assertTrue(subsets[0][0].shape[0] > 395 and \
                        subsets[0][0].shape[0] < 405)
        self.assertTrue(subsets[0][1].shape[0] > 395 and \
                        subsets[0][1].shape[0] < 405)
        self.assertTrue(subsets[1][0].shape[0] > 295 and \
                        subsets[1][0].shape[0] < 305)
        self.assertTrue(subsets[2][0].shape[0] > 45 and \
                        subsets[2][0].shape[0] < 55)
        self.assertTrue(subsets[3][0].shape[0] > 245 and \
                        subsets[3][0].shape[0] < 255)

    # Check metadata sampling via percentage split
    def test_PercentageSplit_metadata(self):
        subsets = sampling_split(self.x, self.y, metadata=self.y,
                               sampling=[0.4, 0.3, 0.05, 0.25],
                               iterative=False, stratified=False)
        self.assertTrue(len(subsets) == 4)
        self.assertTrue(len(subsets[0]) == 3)
        self.assertTrue(len(subsets[1]) == 3)
        self.assertTrue(len(subsets[2]) == 3)
        self.assertTrue(len(subsets[3]) == 3)
        self.assertTrue(subsets[0][0].shape[0] > 395 and \
                      subsets[0][0].shape[0] < 405)
        self.assertTrue(subsets[0][2].shape[0] > 395 and \
                      subsets[0][2].shape[0] < 405)
        self.assertTrue(subsets[1][2].shape[0] > 295 and \
                      subsets[1][2].shape[0] < 305)
        self.assertTrue(subsets[2][2].shape[0] > 45 and \
                      subsets[2][2].shape[0] < 55)
        self.assertTrue(subsets[3][2].shape[0] > 245 and \
                      subsets[3][2].shape[0] < 255)

    #-------------------------------------------------#
    #       Sampling via k-fold Cross-Validation      #
    #-------------------------------------------------#
    # Check random sampling via k-fold cross-validation
    def test_CrossValidation_random(self):
        subsets = sampling_kfold(self.x, self.y, n_splits=5,
                                 iterative=False, stratified=False)
        self.assertEqual(len(subsets), 5)
        for fold in subsets:
            (tx, ty, vx, vy) = fold
            self.assertTrue(tx.shape[0] > 795 and tx.shape[0] < 805)
            self.assertTrue(ty.shape[0] > 795 and ty.shape[0] < 805)
            self.assertTrue(vx.shape[0] > 195 and vx.shape[0] < 205)
            self.assertTrue(vy.shape[0] > 195 and vy.shape[0] < 205)

    # Check stratified random sampling via k-fold cross-validation
    def test_CrossValidation_stratified(self):
        subsets = sampling_kfold(self.x, self.y, n_splits=5,
                                 iterative=False, stratified=True)
        self.assertEqual(len(subsets), 5)
        for fold in subsets:
            (tx, ty, vx, vy) = fold
            self.assertTrue(tx.shape[0] > 795 and tx.shape[0] < 805)
            self.assertTrue(ty.shape[0] > 795 and ty.shape[0] < 805)
            self.assertTrue(vx.shape[0] > 195 and vx.shape[0] < 205)
            self.assertTrue(vy.shape[0] > 195 and vy.shape[0] < 205)

    # Check stratified iterative sampling via k-fold cross-validation
    def test_CrossValidation_iterative(self):
        subsets = sampling_kfold(self.x, self.y, n_splits=5,
                                 iterative=True, stratified=True)
        self.assertEqual(len(subsets), 5)
        for fold in subsets:
            (tx, ty, vx, vy) = fold
            self.assertTrue(tx.shape[0] > 795 and tx.shape[0] < 805)
            self.assertTrue(ty.shape[0] > 795 and ty.shape[0] < 805)
            self.assertTrue(vx.shape[0] > 195 and vx.shape[0] < 205)
            self.assertTrue(vy.shape[0] > 195 and vy.shape[0] < 205)

    # Check metadata sampling via k-fold cross-validation
    def test_CrossValidation_metadata(self):
        subsets = sampling_kfold(self.x, self.y, metadata=self.y, n_splits=5)
        self.assertEqual(len(subsets), 5)
        for fold in subsets:
            (tx, ty, tm, vx, vy, vm) = fold
            self.assertTrue(tx.shape[0] > 795 and tx.shape[0] < 805)
            self.assertTrue(tm.shape[0] > 795 and tm.shape[0] < 805)
            self.assertTrue(vx.shape[0] > 195 and vx.shape[0] < 205)
            self.assertTrue(vm.shape[0] > 195 and vm.shape[0] < 205)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
from tensorflow.keras.callbacks import EarlyStopping
import pandas as pd

#-----------------------------------------------------#
#                   Custom Callbacks                  #
#-----------------------------------------------------#
class ThresholdEarlyStopping(EarlyStopping):
    """ Changed baseline to act as a real baseline.

    The number of patience epochs are only counted when baseline loss is achieved.

    ??? abstract "Reference - Implementation"
        Author:   JBSnorro <br>
        Source:   https://stackoverflow.com/questions/53500047/stop-training-in-keras-when-accuracy-is-already-1-0  <br>
    """
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.baseline_attained = False

    def on_epoch_end(self, epoch, logs=None):
        if not self.baseline_attained:
            current = self.get_monitor_value(logs)
            if current is None:
                return
            if self.monitor_op(current, self.baseline):
                if self.verbose > 0:
                    print('Baseline attained.')
                self.baseline_attained = True
            else:
                return
        super(ThresholdEarlyStopping, self).on_epoch_end(epoch, logs)

class MinEpochEarlyStopping(EarlyStopping):
    """ Changed baseline to act as a real baseline.

    The number of patience epochs are only counted when baseline loss is achieved.

    ??? abstract "Reference - Implementation"
        Author:   McLawrence  <br>
        Source:   https://stackoverflow.com/questions/46287403/is-there-a-way-to-implement-early-stopping-in-keras-only-after-the-first-say-1  <br>
    """
    def __init__(self, monitor='val_loss', min_delta=0, patience=0, verbose=0,
                 mode='auto', start_epoch = 100): # add argument for starting epoch
        super(MinEpochEarlyStopping, self).__init__()
        self.start_epoch = start_epoch

    def on_epoch_end(self, epoch, logs=None):
        if epoch > self.start_epoch:
            super(MinEpochEarlyStopping, self).on_epoch_end(epoch, logs)

#-----------------------------------------------------#
#                    Callback Utils                   #
#-----------------------------------------------------#
def csv_to_history(input_path):
    """ Utility function for reading a CSV file from the
    [CSVLogger](https://www.tensorflow.org/api_docs/python/tf/keras/callbacks/CSVLogger)
    Callback and return a History dictionary object.

    Can be utilized in order to pass returned dictionary object to the
    [evaluate_fitting()][aucmedi.evaluation.fitting] function of the AUCMEDI
    [evaluation][aucmedi.evaluation] submodule.

    Args:
        input_path (str):           Path to a CSV file generated by a CSVLogger Callback.

    Returns:
        history (dict):       A history dictionary from a Keras history object which contains several logs.
    """
    # Read logging data
    dt = pd.read_csv(input_path, sep=",")
    # Parse to dict and return results
    return dt.to_dict(orient="list")



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import os
import json
import pandas as pd
# Internal libraries
from aucmedi import *
from aucmedi.data_processing.io_loader import image_loader, sitk_loader
from aucmedi.data_processing.subfunctions import *
from aucmedi.ensemble import *
from aucmedi.xai import xai_decoder

#-----------------------------------------------------#
#            Building Blocks for Inference            #
#-----------------------------------------------------#
def block_predict(config):
    """ Internal code block for AutoML inference.

    This function is called by the Command-Line-Interface (CLI) of AUCMEDI.

    Args:
        config (dict):                      Configuration dictionary containing all required
                                            parameters for performing an AutoML inference.

    The following attributes are stored in the `config` dictionary:

    Attributes:
        path_imagedir (str):                Path to the directory containing the images for prediction.
        path_modeldir (str):                Path to the model directory in which fitted model weights and metadata are stored.
        path_pred (str):                    Path to the output file in which predicted csv file should be stored.
        xai_method (str or None):           Key for XAI method.
        xai_directory (str or None):        Path to the output directory in which predicted image xai heatmaps should be stored.
        batch_size (int):                   Number of samples inside a single batch.
        workers (int):                      Number of workers/threads which preprocess batches during runtime.
    """
    # Peak into the dataset via the input interface
    ds = input_interface("directory",
                         config["path_imagedir"],
                         path_data=None,
                         training=False,
                         ohe=False,
                         image_format=None)
    (index_list, _, _, _, image_format) = ds

    # Verify existence of input directory
    if not os.path.exists(config["path_modeldir"]):
        raise FileNotFoundError(config["path_modeldir"])

    # Load metadata from training
    path_meta = os.path.join(config["path_modeldir"], "meta.training.json")
    with open(path_meta, "r") as json_file:
        meta_training = json.load(json_file)

    # Define neural network parameters
    nn_paras = {"n_labels": 1,                                  # placeholder
                "channels": 1,                                  # placeholder
    }
    # Select input shape for 3D
    if meta_training["three_dim"]:
        nn_paras["input_shape"] = tuple(meta_training["shape_3D"])

    # Subfunctions
    sf_list = []
    if meta_training["three_dim"]:
        sf_norm = Standardize(mode="grayscale")
        sf_pad = Padding(mode="constant", shape=meta_training["shape_3D"])
        sf_crop = Crop(shape=meta_training["shape_3D"], mode="random")
        sf_chromer = Chromer(target="rgb")
        sf_list.extend([sf_norm, sf_pad, sf_crop, sf_chromer])

    # Define parameters for DataGenerator
    paras_datagen = {
        "path_imagedir": config["path_imagedir"],
        "batch_size": config["batch_size"],
        "img_aug": None,
        "subfunctions": sf_list,
        "prepare_images": False,
        "sample_weights": None,
        "seed": None,
        "image_format": image_format,
        "workers": config["workers"],
        "shuffle": False,
        "grayscale": False,
    }
    if not meta_training["three_dim"] : paras_datagen["loader"] = image_loader
    else : paras_datagen["loader"] = sitk_loader

    # Apply MIC pipelines
    if meta_training["analysis"] == "minimal":
        # Setup neural network
        if not meta_training["three_dim"]:
            arch_dim = "2D." + meta_training["architecture"]
        else : arch_dim = "3D." + meta_training["architecture"]
        model = NeuralNetwork(architecture=arch_dim, **nn_paras)

        # Build DataGenerator
        pred_gen = DataGenerator(samples=index_list,
                                 labels=None,
                                 resize=model.meta_input,
                                 standardize_mode=model.meta_standardize,
                                 **paras_datagen)
        # Load model
        path_model = os.path.join(config["path_modeldir"], "model.last.keras")
        model.load(path_model)
        # Start model inference
        preds = model.predict(prediction_generator=pred_gen)
    elif meta_training["analysis"] == "standard":
        # Setup neural network
        if not meta_training["three_dim"]:
            arch_dim = "2D." + meta_training["architecture"]
        else : arch_dim = "3D." + meta_training["architecture"]
        model = NeuralNetwork(architecture=arch_dim, **nn_paras)

        # Build DataGenerator
        pred_gen = DataGenerator(samples=index_list,
                                 labels=None,
                                 resize=model.meta_input,
                                 standardize_mode=model.meta_standardize,
                                 **paras_datagen)
        # Load model
        path_model = os.path.join(config["path_modeldir"],
                                  "model.best_loss.keras")
        model.load(path_model)
        # Start model inference via Augmenting
        preds = predict_augmenting(model, pred_gen)
    else:
        # Build multi-model list
        model_list = []
        for arch in meta_training["architecture"]:
            if not meta_training["three_dim"] : arch_dim = "2D." + arch
            else : arch_dim = "3D." + arch
            model_part = NeuralNetwork(architecture=arch_dim, **nn_paras)
            model_list.append(model_part)
        el = Composite(model_list, metalearner=meta_training["metalearner"],
                       k_fold=len(meta_training["architecture"]))

        # Build DataGenerator
        pred_gen = DataGenerator(samples=index_list,
                                 labels=None,
                                 resize=None,
                                 standardize_mode=None,
                                 **paras_datagen)
        # Load composite model directory
        el.load(config["path_modeldir"])
        # Start model inference via ensemble learning
        preds = el.predict(pred_gen)

    # Create prediction dataset
    df_index = pd.DataFrame(data={"SAMPLE": index_list})
    df_pd = pd.DataFrame(data=preds, columns=meta_training["class_names"])
    df_merged = pd.concat([df_index, df_pd], axis=1, sort=False)
    df_merged.sort_values(by=["SAMPLE"], inplace=True)
    # Store predictions to disk
    df_merged.to_csv(config["path_pred"], index=False)

    # Create XAI heatmaps
    if config["xai_method"] is not None and config["xai_directory"] is not None:
        if meta_training["analysis"] == "advanced":
            raise ValueError("XAI is only supported for single model pipelines!")
        # Create xai output directory
        if not os.path.exists(config["xai_directory"]):
            os.mkdir(config["xai_directory"])
        # Run XAI decoder
        xai_decoder(pred_gen, model, preds=preds, method=config["xai_method"],
                    layerName=None, alpha=0.4, out_path=config["xai_directory"])



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
from lime import lime_image
# Internal Libraries
from aucmedi.xai.methods.xai_base import XAImethod_Base

#-----------------------------------------------------#
#                  LIME: Pro Features                 #
#-----------------------------------------------------#
class LimePro(XAImethod_Base):
    """ XAI Method for LIME Pro.

    Normally, this class is used internally in the [aucmedi.xai.decoder.xai_decoder][] in the AUCMEDI XAI module.

    ??? abstract "Reference - Implementation"
        Lime: Explaining the predictions of any machine learning classifier <br>
        GitHub: [https://github.com/marcotcr/lime](https://github.com/marcotcr/lime) <br>

    ??? abstract "Reference - Publication"
        Marco Tulio Ribeiro, Sameer Singh, Carlos Guestrin. 9 Aug 2016.
        "Why Should I Trust You?": Explaining the Predictions of Any Classifier
        <br>
        [https://arxiv.org/abs/1602.04938](https://arxiv.org/abs/1602.04938)

    This class provides functionality for running the compute_heatmap function,
    which computes a Lime Pro Map for an image with a model.
    """
    def __init__(self, model, layerName=None, num_samples=1000):
        """ Initialization function for creating a Lime Pro Map as XAI Method object.

        Args:
            model (keras.model):            Keras model object.
            layerName (str):                Not required in LIME Pro/Con Maps, but defined by Abstract Base Class.
            num_samples (int):              Number of iterations for LIME instance explanation.
        """
        # Cache class parameters
        self.model = model
        self.num_samples = num_samples

    #---------------------------------------------#
    #             Heatmap Computation             #
    #---------------------------------------------#
    def compute_heatmap(self, image, class_index, eps=1e-8):
        """ Core function for computing the Lime Pro Map for a provided image and for specific classification outcome.

        ???+ attention
            Be aware that the image has to be provided in batch format.

        Args:
            image (numpy.ndarray):              Image matrix encoded as NumPy Array (provided as one-element batch).
            class_index (int):                  Classification index for which the heatmap should be computed.
            eps (float):                        Epsilon for rounding.

        The returned heatmap is encoded within a range of [0,1]

        ???+ attention
            The shape of the returned heatmap is 2D -> batch and channel axis will be removed.

        Returns:
            heatmap (numpy.ndarray):            Computed Lime Pro Map for provided image.
        """
        # Initialize LIME explainer
        explainer = lime_image.LimeImageExplainer()
        explanation = explainer.explain_instance(image[0].astype("double"),
                                self.model.predict, hide_color=0,
                                labels=(class_index,),
                                num_samples=self.num_samples)
        # Obtain PRO explanation mask
        temp, mask = explanation.get_image_and_mask(class_index, hide_rest=True,
                                positive_only=True, negative_only=False)
        heatmap = mask
        # Intensity normalization to [0,1]
        numer = heatmap - np.min(heatmap)
        denom = (heatmap.max() - heatmap.min()) + eps
        heatmap = numer / denom
        # Return the resulting heatmap
        return heatmap



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
from abc import ABC, abstractmethod

#-----------------------------------------------------#
#         Abstract Base Class for Metalearner         #
#-----------------------------------------------------#
class Metalearner_Base(ABC):
    """ An abstract base class for a Metalearner class.

    Metalearner are similar to [Aggregate functions][aucmedi.ensemble.aggregate],
    however Metalearners are models which require fitting before usage.

    Metalearners are utilized in [Stacking][aucmedi.ensemble.stacking] pipelines.

    A Metalearner act as a combiner algorithm which is trained to make a final prediction
    using predictions of other algorithms (`NeuralNetwork`) as inputs.

    ```
    Assembled predictions encoded in a NumPy matrix with shape (N_models, N_classes).
    Example: [[0.5, 0.4, 0.1],
              [0.4, 0.3, 0.3],
              [0.5, 0.2, 0.3]]
    -> shape (3, 3)

    Merged prediction encoded in a NumPy matrix with shape (1, N_classes).
    Example: [[0.4, 0.3, 0.3]]
    -> shape (1, 3)
    ```

    !!! info "Required Functions"
        | Function            | Description                                                |
        | ------------------- | ---------------------------------------------------------- |
        | `__init__()`        | Object creation function.                                  |
        | `training()`        | Fit Metalearner model.                                     |
        | `prediction()`      | Merge multiple class predictions into a single prediction. |
        | `dump()`            | Store Metalearner model to disk.                           |
        | `load()`            | Load Metalearner model from disk.                          |
    """
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    @abstractmethod
    def __init__(self):
        """ Initialization function which will be called during the Metalearner object creation.

        This function can be used to pass variables and options in the Metalearner instance.
        There are no mandatory parameters for the initialization.
        """
        pass

    #---------------------------------------------#
    #                  Training                   #
    #---------------------------------------------#
    @abstractmethod
    def train(self, x, y):
        """ Training function to fit the Metalearner model.

        Args:
            x (numpy.ndarray):          Assembled prediction dataset encoded in a NumPy matrix with shape (N_samples, N_classes*N_models).
            y (numpy.ndarray):          Classification list with One-Hot Encoding. Provided by
                                        [input_interface][aucmedi.data_processing.io_data.input_interface].
        """
        pass

    #---------------------------------------------#
    #                  Prediction                 #
    #---------------------------------------------#
    @abstractmethod
    def predict(self, data):
        """ Merge multiple predictions for a sample into a single prediction.

        It is required to return the merged predictions (as NumPy matrix).
        It is possible to pass configurations through the initialization function for this class.

        Args:
            data (numpy.ndarray):       Assembled predictions encoded in a NumPy matrix with shape (N_models, N_classes).
        Returns:
            pred (numpy.ndarray):       Merged prediction encoded in a NumPy matrix with shape (1, N_classes).
        """
        pass

    #---------------------------------------------#
    #              Dump Model to Disk             #
    #---------------------------------------------#
    @abstractmethod
    def dump(self, path):
        """ Store Metalearner model to disk.

        Args:
            path (str):                 Path to store the model on disk.
        """
        pass

    #---------------------------------------------#
    #             Load Model from Disk            #
    #---------------------------------------------#
    @abstractmethod
    def load(self, path):
        """ Load Metalearner model and its weights from a file.

        Args:
            path (str):                 Input path from which the model will be loaded.
        """
        pass



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import numpy as np
import tempfile
from PIL import Image
import os
import shutil
#Internal libraries
from aucmedi import DataGenerator
from aucmedi.data_processing.io_loader import numpy_loader

#-----------------------------------------------------#
#               Unittest: Data Generator              #
#-----------------------------------------------------#
class DataGeneratorTEST(unittest.TestCase):
    # Create random imaging and classification data
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Initialize temporary directory
        self.tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                    suffix=".data")
        # Create Grayscale data for 2D
        self.sampleList_gray_2D = []
        for i in range(0, 25):
            img_gray = np.random.rand(16, 16) * 255
            imgGRAY_pillow = Image.fromarray(img_gray.astype(np.uint8))
            index = "image.sample_" + str(i) + ".GRAY.png"
            path_sampleGRAY = os.path.join(self.tmp_data.name, index)
            imgGRAY_pillow.save(path_sampleGRAY)
            self.sampleList_gray_2D.append(index)
        # Create RGB data for 2D
        self.sampleList_rgb_2D = []
        for i in range(0, 25):
            img_rgb = np.random.rand(16, 16, 3) * 255
            imgRGB_pillow = Image.fromarray(img_rgb.astype(np.uint8))
            index = "image.sample_" + str(i) + ".RGB.png"
            path_sampleRGB = os.path.join(self.tmp_data.name, index)
            imgRGB_pillow.save(path_sampleRGB)
            self.sampleList_rgb_2D.append(index)
        # Create Grayscale data for 3D
        self.sampleList_gray_3D = []
        for i in range(0, 25):
            img_gray = np.random.rand(16, 16, 16) * 255
            index = "image.sample_" + str(i) + ".GRAY.npy"
            path_sampleGRAY = os.path.join(self.tmp_data.name, index)
            np.save(path_sampleGRAY, img_gray)
            self.sampleList_gray_3D.append(index)
        # Create RGB data for 3D
        self.sampleList_rgb_3D = []
        for i in range(0, 25):
            img_rgb = np.random.rand(16, 16, 16, 3) * 255
            index = "image.sample_" + str(i) + ".RGB.npy"
            path_sampleRGB = os.path.join(self.tmp_data.name, index)
            np.save(path_sampleRGB, img_rgb)
            self.sampleList_rgb_3D.append(index)

        # Create classification labels
        self.labels_ohe = np.zeros((25, 4), dtype=np.uint8)
        for i in range(0, 25):
            class_index = np.random.randint(0, 4)
            self.labels_ohe[i][class_index] = 1

        # Create metadata
        self.metadata = np.zeros((25, 10), dtype=np.uint8)
        for i in range(0, 25):
            class_index = np.random.randint(0, 10)
            self.metadata[i][class_index] = 1

    #-------------------------------------------------#
    #           Initialization Functionality          #
    #-------------------------------------------------#
    # Class Creation
    def test_BASE_create(self):
        data_gen = DataGenerator(self.sampleList_rgb_2D, self.tmp_data.name)
        self.assertIsInstance(data_gen, DataGenerator)

    #-------------------------------------------------#
    #        Application Functionality for 2D         #
    #-------------------------------------------------#
    # Usage: Grayscale without Labels
    def test_RUN_2D_GRAYSCALE_noLabel(self):
        data_gen = DataGenerator(self.sampleList_gray_2D, self.tmp_data.name,
                                 grayscale=True, batch_size=5)
        for i in range(0, 10):
            batch = data_gen[i]
            self.assertTrue(len(batch), 1)
            self.assertTrue(np.array_equal(batch[0].shape, (5, 224, 224, 1)))

    # Usage: RGB without Labels
    def test_RUN_2D_RGB_noLabel(self):
        data_gen = DataGenerator(self.sampleList_rgb_2D, self.tmp_data.name,
                                 grayscale=False, batch_size=5)
        for i in range(0, 10):
            batch = data_gen[i]
            self.assertTrue(len(batch), 1)
            self.assertTrue(np.array_equal(batch[0].shape, (5, 224, 224, 3)))

    # Usage: With Labels
    def test_RUN_2D_withLabel(self):
        data_gen = DataGenerator(self.sampleList_rgb_2D, self.tmp_data.name,
                                 labels=self.labels_ohe,
                                 grayscale=False, batch_size=5)
        for i in range(0, 10):
            batch = data_gen[i]
            self.assertTrue(len(batch), 2)
            self.assertTrue(np.array_equal(batch[1].shape, (5, 4)))

    #-------------------------------------------------#
    #        Application Functionality for 3D         #
    #-------------------------------------------------#
    # Usage: Grayscale without Labels
    def test_RUN_3D_GRAYSCALE_noLabel(self):
        data_gen = DataGenerator(self.sampleList_gray_3D, self.tmp_data.name,
                                 grayscale=True, batch_size=5, two_dim=False,
                                 loader=numpy_loader, resize=None,
                                 standardize_mode=None)
        for i in range(0, 10):
            batch = data_gen[i]
            self.assertTrue(len(batch), 1)
            self.assertTrue(np.array_equal(batch[0].shape, (5, 16, 16, 16, 1)))

    # Usage: RGB without Labels
    def test_RUN_3D_RGB_noLabel(self):
        data_gen = DataGenerator(self.sampleList_rgb_3D, self.tmp_data.name,
                                 grayscale=False, batch_size=5, two_dim=False,
                                 loader=numpy_loader, resize=None,
                                 standardize_mode=None)
        for i in range(0, 10):
            batch = data_gen[i]
            self.assertTrue(len(batch), 1)
            self.assertTrue(np.array_equal(batch[0].shape, (5, 16, 16, 16, 3)))

    # Usage: With Labels
    def test_RUN_3D_withLabel(self):
        data_gen = DataGenerator(self.sampleList_rgb_3D, self.tmp_data.name,
                                 labels=self.labels_ohe, two_dim=False,
                                 grayscale=False, batch_size=5,
                                 loader=numpy_loader, resize=None,
                                 standardize_mode=None)
        for i in range(0, 10):
            batch = data_gen[i]
            self.assertTrue(len(batch), 2)
            self.assertTrue(np.array_equal(batch[1].shape, (5, 4)))

    #-------------------------------------------------#
    #     Application Functionality with Metadata     #
    #-------------------------------------------------#
    # Usage: Metadata for inference
    def test_RUN_Metadata_noLabel(self):
        data_gen = DataGenerator(self.sampleList_rgb_2D, self.tmp_data.name,
                                 metadata=self.metadata, grayscale=False,
                                 batch_size=5)
        for i in range(0, 10):
            batch = data_gen[i]
            self.assertTrue(len(batch), 1)
            self.assertTrue(len(batch[0]) == 2)
            self.assertTrue(np.array_equal(batch[0][0].shape, (5, 224, 224, 3)))
            self.assertTrue(np.array_equal(batch[0][1].shape, (5, 10)))

    # Usage: Metadata for training
    def test_RUN_Metadata_withLabel(self):
        data_gen = DataGenerator(self.sampleList_rgb_2D, self.tmp_data.name,
                             labels=self.labels_ohe, metadata=self.metadata,
                             grayscale=False, batch_size=5)
        for i in range(0, 10):
            batch = data_gen[i]
            self.assertTrue(len(batch), 2)
            self.assertTrue(np.array_equal(batch[1].shape, (5, 4)))
            self.assertTrue(len(batch[0]) == 2)
            self.assertTrue(np.array_equal(batch[0][0].shape, (5, 224, 224, 3)))
            self.assertTrue(np.array_equal(batch[0][1].shape, (5, 10)))

    #-------------------------------------------------#
    #                 Multi-Processing                #
    #-------------------------------------------------#
    def test_MP(self):
        data_gen = DataGenerator(self.sampleList_rgb_2D, self.tmp_data.name,
                                 labels=self.labels_ohe,
                                 grayscale=False, batch_size=5, workers=5)
        for i in range(0, 10):
            batch = data_gen[i]
            self.assertTrue(len(batch), 2)
            self.assertTrue(np.array_equal(batch[1].shape, (5, 4)))

    #-------------------------------------------------#
    #             Beforehand Preprocessing            #
    #-------------------------------------------------#
    def test_PrepareImages(self):
        data_gen = DataGenerator(self.sampleList_rgb_2D, self.tmp_data.name,
                                 labels=self.labels_ohe, prepare_images=True,
                                 grayscale=False, batch_size=5)
        precprocessed_images = os.listdir(data_gen.prepare_dir)
        self.assertTrue(len(precprocessed_images), len(self.sampleList_rgb_2D))
        for i in range(0, 10):
            batch = data_gen[i]
            self.assertTrue(len(batch), 2)
            self.assertTrue(np.array_equal(batch[1].shape, (5, 4)))
        shutil.rmtree(data_gen.prepare_dir)

    def test_PrepareImages_MP(self):
        data_gen = DataGenerator(self.sampleList_rgb_2D, self.tmp_data.name,
                                 labels=self.labels_ohe, prepare_images=True,
                                 grayscale=False, batch_size=5, workers=5)
        precprocessed_images = os.listdir(data_gen.prepare_dir)
        self.assertTrue(len(precprocessed_images), len(self.sampleList_rgb_2D))
        for i in range(0, 10):
            batch = data_gen[i]
            self.assertTrue(len(batch), 2)
            self.assertTrue(np.array_equal(batch[1].shape, (5, 4)))
        shutil.rmtree(data_gen.prepare_dir)

    #-------------------------------------------------#
    #                   Utilization                   #
    #-------------------------------------------------#
    # Class Creation
    # def test_utils_iter(self):
    #    data_gen = DataGenerator(self.sampleList_rgb_2D, self.tmp_data.name,
    #                             batch_size=8)
    #    counter = 0
    #    for batch in data_gen:
    #        if counter < 3:
    #            self.assertTrue(np.array_equal(batch[0].shape, (8,224,224,3)))
    #        else: 
    #            self.assertTrue(np.array_equal(batch[0].shape, (1,224,224,3)))
    #        counter += 1
    #    self.assertTrue(counter == 4)


#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
import tensorflow as tf
# Internal Libraries
from aucmedi.xai.methods.xai_base import XAImethod_Base

#-----------------------------------------------------#
#           Saliency Maps / Backpropagation           #
#-----------------------------------------------------#
class SaliencyMap(XAImethod_Base):
    """ XAI Method for Saliency Map (also called Backpropagation).

    Normally, this class is used internally in the [aucmedi.xai.decoder.xai_decoder][] in the AUCMEDI XAI module.

    ??? abstract "Reference - Implementation #1"
        Author: Yasuhiro Kubota <br>
        GitHub Profile: [https://github.com/keisen](https://github.com/keisen) <br>
        Date: Aug 11, 2020 <br>
        [https://github.com/keisen/tf-keras-vis/](https://github.com/keisen/tf-keras-vis/) <br>

    ??? abstract "Reference - Implementation #2"
        Author: Huynh Ngoc Anh <br>
        GitHub Profile: [https://github.com/experiencor](https://github.com/experiencor) <br>
        Date: Jun 23, 2017 <br>
        [https://github.com/experiencor/deep-viz-keras/](https://github.com/experiencor/deep-viz-keras/) <br>

    ??? abstract "Reference - Publication"
        Karen Simonyan, Andrea Vedaldi, Andrew Zisserman. 20 Dec 2013.
        Deep Inside Convolutional Networks: Visualising Image Classification Models and Saliency Maps.
        <br>
        [https://arxiv.org/abs/1312.6034](https://arxiv.org/abs/1312.6034)

    This class provides functionality for running the compute_heatmap function,
    which computes a Saliency Map for an image with a model.
    """
    def __init__(self, model, layerName=None):
        """ Initialization function for creating a Saliency Map as XAI Method object.

        Args:
            model (keras.model):               Keras model object.
            layerName (str):                   Not required in Saliency Maps, but defined by Abstract Base Class.
        """
        # Cache class parameters
        self.model = model

    #---------------------------------------------#
    #             Heatmap Computation             #
    #---------------------------------------------#
    def compute_heatmap(self, image, class_index, eps=1e-8):
        """ Core function for computing the Saliency Map for a provided image and for specific classification outcome.

        ???+ attention
            Be aware that the image has to be provided in batch format.

        Args:
            image (numpy.ndarray):              Image matrix encoded as NumPy Array (provided as one-element batch).
            class_index (int):                  Classification index for which the heatmap should be computed.
            eps (float):                        Epsilon for rounding.

        The returned heatmap is encoded within a range of [0,1]

        ???+ attention
            The shape of the returned heatmap is 2D -> batch and channel axis will be removed.

        Returns:
            heatmap (numpy.ndarray):            Computed Saliency Map for provided image.
        """
        # Compute gradient for desierd class index
        with tf.GradientTape() as tape:
            inputs = tf.cast(image, tf.float32)
            tape.watch(inputs)
            preds = self.model(inputs)
            loss = preds[:, class_index]
        gradient = tape.gradient(loss, inputs)
        # Obtain maximum gradient based on feature map of last conv layer
        gradient = tf.reduce_max(gradient, axis=-1)
        # Convert to NumPy & Remove batch axis
        heatmap = gradient.numpy()[0,:,:]

        # Intensity normalization to [0,1]
        numer = heatmap - np.min(heatmap)
        denom = (heatmap.max() - heatmap.min()) + eps
        heatmap = numer / denom

        # Return the resulting heatmap
        return heatmap



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import pickle
from sklearn.metrics import roc_auc_score
import numpy as np
# Internal libraries/scripts
from aucmedi.ensemble.metalearner.ml_base import Metalearner_Base

#-----------------------------------------------------#
#              Metalearner: Best Model                #
#-----------------------------------------------------#
class BestModel(Metalearner_Base):
    """ A Best Model based Metalearner.

    This class should be passed to an ensemble function/class like Stacking for combining predictions.

    This Metalearner computes the Area Under the Receiver Operating Characteristic Curve (ROC AUC)
    for each model, and simply utilizes only the predictions of the best scoring model.

    !!! info
        Can be utilized for binary, multi-class and multi-label tasks.
    """
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    def __init__(self):
        self.model = {}

    #---------------------------------------------#
    #                  Training                   #
    #---------------------------------------------#
    def train(self, x, y):
        # Identify number of models and classes
        n_classes = y.shape[1]
        n_models = int(x.shape[1] / n_classes)
        # Preprocess data input
        data = np.reshape(x, (x.shape[0], n_models, n_classes))

        # Compute AUC scores and store them to cache
        for m in range(n_models):
            pred = data[:,m,:]
            score = roc_auc_score(y, pred, average="macro")
            self.model["nn_" + str(m)] = score

        # Identify best model and store results to cache
        best_model = max(self.model, key=self.model.get)
        self.model["best_model"] = best_model
        self.model["n_classes"] = n_classes
        self.model["n_models"] = n_models

    #---------------------------------------------#
    #                  Prediction                 #
    #---------------------------------------------#
    def predict(self, data):
        # Preprocess data input
        preds = np.reshape(data, (data.shape[0],
                                  self.model["n_models"],
                                  self.model["n_classes"]))

        # Obtain prediction probabilities of best model
        m = int(self.model["best_model"].split("_")[-1])
        pred_best = preds[:,m,:]
        # Return results
        return pred_best

    #---------------------------------------------#
    #              Dump Model to Disk             #
    #---------------------------------------------#
    def dump(self, path):
        # Dump model to disk via pickle
        with open(path, "wb") as pickle_writer:
            pickle.dump(self.model, pickle_writer)

    #---------------------------------------------#
    #             Load Model from Disk            #
    #---------------------------------------------#
    def load(self, path):
        # Load model from disk via pickle
        with open(path, "rb") as pickle_reader:
            self.model = pickle.load(pickle_reader)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import os
import numpy as np
import json
import pandas as pd

#-----------------------------------------------------#
#         Data Loader Interface based on JSON         #
#-----------------------------------------------------#
def json_loader(path_data, path_imagedir, allowed_image_formats, training=True,
                ohe=True):
    """ Data Input Interface for loading a dataset via a JSON and an image directory.

    This **internal** function allows simple parsing of class annotations encoded in a JSON.

    ???+ info "Input Formats"
        ```
        Format Sparse:
            - Name Index (key) : Class (value)

        Format One-Hot Encoded:
            - Name Index (key) : List consisting of binary integers.
        ```

    **Expected structure:**
    ```
    dataset/
        images_dir/                 # path_imagedir = "dataset/images_dir"
            sample001.png
            sample002.png
            ...
            sample350.png
        annotations.json            # path_data = "dataset/annotations.json"
    ```

    Args:
        path_data (str):                        Path to the json file.
        path_imagedir (str):                    Path to the directory containing the images.
        allowed_image_formats (list of str):    List of allowed imaging formats. (provided by IO_Interface)
        training (bool):                        Boolean option whether annotation data is available.
        ohe (bool):                             Boolean option whether annotation data is sparse categorical or one-hot encoded.

    Returns:
        index_list (list of str):               List of sample/index encoded as Strings. Required in DataGenerator as `samples`.
        class_ohe (numpy.ndarray):              Classification list as One-Hot encoding. Required in DataGenerator as `labels`.
        class_n (int):                          Number of classes. Required in NeuralNetwork for Architecture design as `n_labels`.
        class_names (list of str):              List of names for corresponding classes. Used for later prediction storage or evaluation.
        image_format (str):                     Image format to add at the end of the sample index for image loading. Required in DataGenerator.
    """
    # Load JSON file
    with open(path_data, "r") as json_reader:
        dt_json = json.load(json_reader)
    # Identify image format by peaking first image
    image_format = None
    for file in os.listdir(path_imagedir):
        format = file.split(".")[-1]
        if format.lower() in allowed_image_formats or \
           format.upper() in allowed_image_formats:
           image_format = format
           break
    # Raise Exception if image format is unknown
    if image_format is None:
        raise Exception("Unknown image format.", path_imagedir)

    # Verify if all images are existing
    lever = True
    for sample in dt_json:
        if sample == "legend" : continue
        # Check if image ending is already in sample name by peaking first one
        if lever:
            lever = False
            if sample.endswith("." + image_format) : image_format = None
        # Obtain image file path
        if image_format : img_file = sample + "." + image_format
        else : img_file = sample
        path_img = os.path.join(path_imagedir, img_file)
        # Check existance
        if not os.path.exists(path_img):
            raise Exception("Image does not exist / not accessible!",
                            'Sample: "' + sample + '"', path_img)

    # If JSON is for inference (no annotation data)
    if not training:
        # Ensure index list to contain strings
        if "legend" in dt_json : del dt_json["legend"]
        index_list = [str(x) for x in dt_json]
        # -> return parsing
        return index_list, None, None, None, image_format

    # Try parsing with a sparse categorical class format
    if not ohe:
        # Parse class name information
        if "legend" in dt_json:
            class_names = dt_json["legend"]
            del dt_json["legend"]
        else : class_names = None
        # Obtain class information and index list
        index_list = []
        classes_sparse = []
        for sample in dt_json:
            index_list.append(str(sample))
            classes_sparse.append(dt_json[sample])
        if class_names is None : class_names = np.unique(classes_sparse).tolist()
        class_n = len(class_names)
        # Parse sparse categorical annotations to One-Hot Encoding
        class_ohe = pd.get_dummies(classes_sparse).to_numpy()
    # Try parsing one-hot encoded format
    else:
        # Parse information
        if "legend" in dt_json:
            class_names = dt_json["legend"]
            del dt_json["legend"]
            class_n = len(class_names)
        else:
            class_names = None
            class_n = None
        # Obtain class information and index list
        index_list = []
        class_data = []
        for sample in dt_json:
            index_list.append(str(sample))
            class_data.append(dt_json[sample])
        class_ohe = np.array(class_data)
        # Verify number of class annotation
        if class_n is None : class_ohe.shape[1]

    # Validate if number of samples and number of annotations match
    if len(index_list) != len(class_ohe):
        raise Exception("Numbers of samples and annotations do not match!",
                        len(index_list), len(class_ohe))

    # Return parsed JSON data
    return index_list, class_ohe, class_n, class_names, image_format



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import numpy as np

#-----------------------------------------------------#
#             Cache Loader for AUCMEDI IO             #
#-----------------------------------------------------#
def cache_loader(sample, path_imagedir=None, image_format=None,
                 grayscale=False, two_dim=True, cache=None, **kwargs):
    """ Cache Loader for passing already loaded images within the AUCMEDI pipeline.

    The Cache Loader is an IO_loader function, which have to be passed to the 
    [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].

    The complete data management happens in the memory.
    Thus, for multiple images or common data set sizes, this is NOT recommended!

    !!! warning
        This functions requires to pass a Dictionary to the parameter `cache`!

    Dictionary structure: key=index as String; value=Image as NumPy array <br>
    e.g. cache = {"my_index_001": my_image}

    ???+ example
        ```python
        # Import required libraries
        from aucmedi import *
        from aucmedi.data_processing.io_loader import cache_loader

        # Encode information as dictionary
        cache = {"sample_a": image_a,
                 "sample_b": image_b,
                 "sample_n": image_n}

        # Obtain meta data
        my_labels = [[1,0], [0,1], [1,0]]      # one-hot encoded annotation matrix
        sample_list = cache.keys()

        # Initialize DataGenerator
        data_gen = DataGenerator(sample_list, None, labels=my_labels,
                                 resize=None, grayscale=False, two_dim=True,
                                 loader=cache_loader, cache=cache)
        ```

    Args:
        sample (str):               Sample name/index of an image.
        path_imagedir (str):        Path to the directory containing the images.
        image_format (str):         Image format to add at the end of the sample index for image loading.
        grayscale (bool):           Boolean, whether images are grayscale or RGB.
        two_dim (bool):             Boolean, whether image is 2D or 3D.
        cache (dict):               A Python dictionary containing one or multiple images.
        **kwargs (dict):            Additional parameters for the sample loader.
    """
    # Verify if a cache is provided
    if cache is None or type(cache) is not dict:
        raise TypeError("No dictionary was provided to cache_loader()!")
    # Obtain image from cache
    img = cache[sample]
    # Verify image shape for grayscale & 2D
    if grayscale and two_dim:
        # Add channel axis and return image
        if len(img.shape) == 2:
            return np.reshape(img, img.shape + (1,))
        # Just return image
        elif len(img.shape) == 3 and img.shape[-1] == 1:
            return img
        # Throw Exception
        else:
            raise ValueError("Parameter 2D & Grayscale: Expected either 2D " + \
                             "without channel axis or 3D with single channel" + \
                             " axis, but got:", img.shape, len(img.shape))
    # Verify image shape for grayscale & 3D
    elif grayscale and not two_dim:
        # Add channel axis and return image
        if len(img.shape) == 3:
            return np.reshape(img, img.shape + (1,))
        # Just return image
        elif len(img.shape) == 4 and img.shape[-1] == 1:
            return img
        # Throw Exception
        else:
            raise ValueError("Parameter 3D & Grayscale: Expected either 3D " + \
                             "without channel axis or 4D with single channel" + \
                             " axis, but got:", img.shape, len(img.shape))
    # Verify image shape for rgb & 2D
    elif not grayscale and two_dim:
        # Just return image
        if len(img.shape) == 3 and img.shape[-1] == 3:
            return img
        # Throw Exception
        else:
            raise ValueError("Parameter 2D & RGB: Expected 3D array " + \
                             "including a single channel axis, but got:",
                             img.shape, len(img.shape))
    # Verify image shape for rgb & 3D
    elif not grayscale and not two_dim:
        # Just return image
        if len(img.shape) == 4 and img.shape[-1] == 3:
            return img
        # Throw Exception
        else:
            raise ValueError("Parameter 3D & RGB: Expected 4D array " + \
                             "including a single channel axis, but got:",
                             img.shape, len(img.shape))



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" Entry script (runner/main function) which will be called in AUCMEDI AutoML.

The console entry `aucmedi` refers to `aucmedi.automl.main:main`.

Executes AUCMEDI AutoML pipeline for training, prediction and evaluation.

More information can be found in the docs: [Documentation - AutoML](../../../automl/overview/)
"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import sys
# Internal libraries
from aucmedi.automl import *
from aucmedi.automl.cli import *

#-----------------------------------------------------#
#                Main Method - Runner                 #
#-----------------------------------------------------#
def main():
    # Initialize argparser core
    parser, subparsers = cli_core()
    # # Define Subparser YAML
    # cli_yaml(subparsers)
    # Define Subparser Training
    cli_training(subparsers)
    # Define Subparser Prediction
    cli_prediction(subparsers)
    # Define Subparser Evaluation
    cli_evaluation(subparsers)

    # Help page hook for passing no parameters
    if len(sys.argv)<=1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    # Parse arguments
    else : args = parser.parse_args()

    # Call corresponding cli or yaml parser
    if args.hub == "yaml" : config = parse_yaml(args)
    else : config = parse_cli(args)

    # Run training pipeline
    if config["hub"] == "training" : block_train(config)
    # Run prediction pipeline
    if config["hub"] == "prediction" : block_predict(config)
    # Run evaluation pipeline
    if config["hub"] == "evaluation" : block_evaluate(config)

# Runner for direct script call
if __name__ == "__main__":
    main()



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" API reference for the AUCMEDI AutoML pipeline.

The mentality behind AutoML is to ensure easy application, integration and
maintenance of complex medical image classification pipelines.
AUCMEDI provides a fast and intuitive interface through AUCMEDI AutoML for building,
application and sharing of state-of-the-art medical image classification models.

The AutoML pipelines are categorized into the following modes:
`training`, `prediction` and `evaluation`.

- The console entry `aucmedi` refers to [aucmedi.automl.main:main][aucmedi.automl.main].
- The Argparse interface for CLI is defined in [aucmedi.automl.cli][aucmedi.automl.cli]
- Each AutoML mode is implemented as a code block defining the AUCMEDI pipeline.

!!! info
    | AutoML Mode   | Argparse                                              | Code Block (Pipeline)                         |
    | ------------- | ----------------------------------------------------- | --------------------------------------------- |
    | `training`    | [CLI - Training][aucmedi.automl.cli.cli_training]     | [Block - Train][aucmedi.automl.block_train]   |
    | `prediction`  | [CLI - Prediction][aucmedi.automl.cli.cli_prediction] | [Block - Predict][aucmedi.automl.block_pred]  |
    | `evaluation`  | [CLI - Evaluation][aucmedi.automl.cli.cli_evaluation] | [Block - Evaluate][aucmedi.automl.block_eval] |

More information can be found in the docs: [Documentation - AutoML](../../automl/overview/)
"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# Block functions
from aucmedi.automl.block_train import block_train
from aucmedi.automl.block_pred import block_predict
from aucmedi.automl.block_eval import block_evaluate
# Parser
from aucmedi.automl.parser_yaml import parse_yaml
from aucmedi.automl.parser_cli import parse_cli



from setuptools import setup
from setuptools import find_packages


with open("docs/README.PyPI.md", "r") as fh:
    long_description = fh.read()

setup(
    name='aucmedi',
    version='0.10.0',
    description='AUCMEDI - a framework for Automated Classification of Medical Images',
    author='Dominik Müller',
    author_email='dominik.mueller@informatik.uni-augsburg.de',
    license='GPLv3',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://frankkramer-lab.github.io/aucmedi/',
    project_urls={
        "Bug Tracker": "https://github.com/frankkramer-lab/aucmedi/issues",
        "Documentation": "https://frankkramer-lab.github.io/aucmedi/reference/",
        "Source Code": "https://github.com/frankkramer-lab/aucmedi",
    },
    packages=find_packages(),
    entry_points={
        'console_scripts': ['aucmedi = aucmedi.automl.main:main'],
    },
    python_requires='>=3.9',
    install_requires=['tensorflow>=2.14.0',
                      'numpy>=1.23.5',
                      'pillow>=10.2.0',
                      'albumentations>=1.3.0',
                      'pandas>=1.5.2',
                      'scikit-learn>=1.3.0',
                      'scikit-image>=0.21.0',
                      'lime>=0.2.0.1',
                      'pooch>=1.6.0',
                      'classification-models-3D>=1.0.10',
                      # 'vit-keras>=0.1.2',
                      'Keras-Applications==1.0.8',
                      'SimpleITK>=2.2.0',
                      'batchgenerators>=0.25',
                      'volumentations-aucmedi>=1.0.1',
                      'plotnine==0.12.4',
                      'pathos>=0.3.0'],
    classifiers=["Programming Language :: Python :: 3",
                 "Programming Language :: Python :: 3.10",
                 "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                 "Operating System :: OS Independent",

                 "Intended Audience :: Healthcare Industry",
                 "Intended Audience :: Science/Research",

                 "Topic :: Scientific/Engineering :: Artificial Intelligence",
                 "Topic :: Scientific/Engineering :: Image Recognition",
                 "Topic :: Scientific/Engineering :: Medical Science Apps."]
    )



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import pickle
from sklearn.metrics import roc_auc_score
import numpy as np
# Internal libraries/scripts
from aucmedi.ensemble.metalearner.ml_base import Metalearner_Base

#-----------------------------------------------------#
#             Metalearner: Weighted Mean              #
#-----------------------------------------------------#
class AveragingWeightedMean(Metalearner_Base):
    """ A Weighted Mean based Metalearner.

    This class should be passed to an ensemble function/class like Stacking for combining predictions.

    This Metalearner computes the Area Under the Receiver Operating Characteristic Curve (ROC AUC)
    for each model, and utilizes these scores for a weighted Mean to average predictions.

    !!! info
        Can be utilized for binary, multi-class and multi-label tasks.
    """
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    def __init__(self):
        self.model = {}

    #---------------------------------------------#
    #                  Training                   #
    #---------------------------------------------#
    def train(self, x, y):
        # Identify number of models and classes
        n_classes = y.shape[1]
        n_models = int(x.shape[1] / n_classes)
        # Preprocess data input
        data = np.reshape(x, (x.shape[0], n_models, n_classes))

        # Compute AUC scores and store them to cache
        weights = []
        for m in range(n_models):
            pred = data[:,m,:]
            score = roc_auc_score(y, pred, average="macro")
            weights.append(score)

        # Store results to cache
        self.model["weights"] = weights
        self.model["n_classes"] = n_classes
        self.model["n_models"] = n_models

    #---------------------------------------------#
    #                  Prediction                 #
    #---------------------------------------------#
    def predict(self, data):
        # Preprocess data input
        preds = np.reshape(data, (data.shape[0],
                                  self.model["n_models"],
                                  self.model["n_classes"]))
        # Compute weighted mean
        pred = np.average(preds, axis=1, weights=self.model["weights"])
        # Return results
        return pred

    #---------------------------------------------#
    #              Dump Model to Disk             #
    #---------------------------------------------#
    def dump(self, path):
        # Dump model to disk via pickle
        with open(path, "wb") as pickle_writer:
            pickle.dump(self.model, pickle_writer)

    #---------------------------------------------#
    #             Load Model from Disk            #
    #---------------------------------------------#
    def load(self, path):
        # Load model from disk via pickle
        with open(path, "rb") as pickle_reader:
            self.model = pickle.load(pickle_reader)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" The classification variant of the Vision Transformer (ViT) version L32 architecture.

!!! warning
    The ViT architectures only work for RGB encoding (channel size = 3).

| Architecture Variable    | Value                      |
| ------------------------ | -------------------------- |
| Key in architecture_dict | "2D.ViT_L32"               |
| Input_shape              | (384, 384)                 |
| Standardization          | "tf"                       |

???+ abstract "Reference - Implementation"
    Fausto Morales; [https://github.com/faustomorales](https://github.com/faustomorales) <br>
    [https://github.com/faustomorales/vit-keras](https://github.com/faustomorales/vit-keras) <br>

    Vo Van Tu; [https://github.com/tuvovan](https://github.com/tuvovan) <br>
    [https://github.com/tuvovan/Vision_Transformer_Keras](https://github.com/tuvovan/Vision_Transformer_Keras) <br>

    Original: Google Research <br>
    [https://github.com/google-research/vision_transformer](https://github.com/google-research/vision_transformer) <br>

???+ abstract "Reference - Publication"
    ```
    @article{dosovitskiy2020vit,
      title={An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale},
      author={Dosovitskiy, Alexey and Beyer, Lucas and Kolesnikov, Alexander and Weissenborn, Dirk and Zhai, Xiaohua and Unterthiner, Thomas and  Dehghani, Mostafa and Minderer, Matthias and Heigold, Georg and Gelly, Sylvain and Uszkoreit, Jakob and Houlsby, Neil},
      journal={ICLR},
      year={2021}
    }

    @article{tolstikhin2021mixer,
      title={MLP-Mixer: An all-MLP Architecture for Vision},
      author={Tolstikhin, Ilya and Houlsby, Neil and Kolesnikov, Alexander and Beyer, Lucas and Zhai, Xiaohua and Unterthiner, Thomas and Yung, Jessica and Steiner, Andreas and Keysers, Daniel and Uszkoreit, Jakob and Lucic, Mario and Dosovitskiy, Alexey},
      journal={arXiv preprint arXiv:2105.01601},
      year={2021}
    }

    @article{steiner2021augreg,
      title={How to train your ViT? Data, Augmentation, and Regularization in Vision Transformers},
      author={Steiner, Andreas and Kolesnikov, Alexander and and Zhai, Xiaohua and Wightman, Ross and Uszkoreit, Jakob and Beyer, Lucas},
      journal={arXiv preprint arXiv:2106.10270},
      year={2021}
    }

    @article{chen2021outperform,
      title={When Vision Transformers Outperform ResNets without Pretraining or Strong Data Augmentations},
      author={Chen, Xiangning and Hsieh, Cho-Jui and Gong, Boqing},
      journal={arXiv preprint arXiv:2106.01548},
      year={2021},
    }

    @article{zhai2022lit,
      title={LiT: Zero-Shot Transfer with Locked-image Text Tuning},
      author={Zhai, Xiaohua and Wang, Xiao and Mustafa, Basil and Steiner, Andreas and Keysers, Daniel and Kolesnikov, Alexander and Beyer, Lucas},
      journal={CVPR},
      year={2022}
    }
    ```
"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
from vit_keras import vit
# Internal libraries
from aucmedi.neural_network.architectures import Architecture_Base

#-----------------------------------------------------#
#     Architecture class: Vision Transformer (ViT)    #
#-----------------------------------------------------#
class ViT_L32(Architecture_Base):
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    def __init__(self, classification_head, channels, input_shape=(384, 384),
                 pretrained_weights=False):
        self.classifier = classification_head
        self.input = input_shape + (channels,)
        self.pretrained_weights = pretrained_weights

    #---------------------------------------------#
    #                Create Model                 #
    #---------------------------------------------#
    def create_model(self):
        # Get pretrained image weights from imagenet if desired
        if self.pretrained_weights : pretrained = True
        else : pretrained = False

        # Obtain ViT L32 as base model
        base_model = vit.vit_l32(image_size=self.input[:-1],
                                 classes=self.classifier.n_labels,
                                 include_top=False,
                                 pretrained=pretrained,
                                 pretrained_top=False)
        top_model = base_model.output

        # Add classification head
        model = self.classifier.build(model_input=base_model.input,
                                      model_output=top_model)

        # Return created model
        return model



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" Models are represented with the open-source neural network library [TensorFlow.Keras](https://www.tensorflow.org/api_docs/python/tf/keras)
    which provides an user-friendly API for commonly used neural-network building blocks.

The already implemented architectures are configurable by custom input sizes, optional dropouts, transfer learning via pretrained weights,
meta data inclusion or activation output depending on classification type.

Additionally, AUCMEDI offers architectures for 2D image and 3D volume classification.

???+ example "Example: How to select an Architecture"
    For architecture selection, just create a key (str) by adding "2D." or "3D." to the architecture name,
    and pass the key to the `architecture` parameter of the [NeuralNetwork][aucmedi.neural_network.model.NeuralNetwork] class.

    ```python
    # 2D architecture
    my_model_a = NeuralNetwork(n_labels=8, channels=3, architecture="2D.DenseNet121")
    # 3D architecture for multi-label classification (sigmoid activation)
    my_model_b = NeuralNetwork(n_labels=8, channels=3, architecture="3D.ResNet50",
                                activation_output="sigmoid")
    # 2D architecture with custom input_shape
    my_model_c = NeuralNetwork(n_labels=8, channels=3, architecture="2D.Xception",
                                input_shape=(512,512))
    ```

???+ note "List of implemented Architectures"
    AUCMEDI provides a large library of state-of-the-art and ready-to-use architectures.

    - 2D Architectures: [aucmedi.neural_network.architectures.image][]
    - 3D Architectures: [aucmedi.neural_network.architectures.volume][]

Besides the flexibility in switching between already implemented architecture,
the [abstract base class interface][aucmedi.neural_network.architectures.arch_base.Architecture_Base]
for architectures offers the possibility for custom architecture integration into the AUCMEDI pipeline.

Furthermore, AUCMEDI offers the powerful classification head interface [Classifier][aucmedi.neural_network.architectures.classifier],
which can be used for all types of image classifications and will be automatically created in the
[NeuralNetwork][aucmedi.neural_network.model.NeuralNetwork] class.
"""
#-----------------------------------------------------#
#               General library imports               #
#-----------------------------------------------------#
# Abstract Base Class for Architectures
from aucmedi.neural_network.architectures.arch_base import Architecture_Base
# Classification Head
from aucmedi.neural_network.architectures.classifier import Classifier

#-----------------------------------------------------#
#       Access Functions to Architecture Classes      #
#-----------------------------------------------------#
# Initialize combined architecture_dict for image & volume architectures
architecture_dict = {}

# Add image architectures to architecture_dict
from aucmedi.neural_network.architectures.image import architecture_dict as arch_image
for arch in arch_image:
    architecture_dict["2D." + arch] = arch_image[arch]

# Add volume architectures to architecture_dict
from aucmedi.neural_network.architectures.volume import architecture_dict as arch_volume
for arch in arch_volume:
    architecture_dict["3D." + arch] = arch_volume[arch]

#-----------------------------------------------------#
#       Meta Information of Architecture Classes      #
#-----------------------------------------------------#
# Initialize combined supported_standardize_mode for image & volume architectures
supported_standardize_mode = {}

# Add image architectures to supported_standardize_mode
from aucmedi.neural_network.architectures.image import supported_standardize_mode as modes_image
for m in modes_image:
    supported_standardize_mode["2D." + m] = modes_image[m]

# Add volume architectures to supported_standardize_mode
from aucmedi.neural_network.architectures.volume import supported_standardize_mode as modes_volume
for m in modes_volume:
    supported_standardize_mode["3D." + m] = modes_volume[m]



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" This is the API reference for the AUCMEDI framework.

Build your state-of-the-art medical image classification pipeline with the 3 AUCMEDI pillars:

!!! info "Pillars of AUCMEDI"
    | Pillar                                                                    | Description                                                       |
    | ------------------------------------------------------------------------- | ----------------------------------------------------------------- |
    | #1: [input_interface()][aucmedi.data_processing.io_data.input_interface]  | Obtaining general information from the dataset.                   |
    | #2: [NeuralNetwork][aucmedi.neural_network.model.NeuralNetwork]           | Building the deep learning model.                                 |
    | #3: [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator] | Powerful interface for loading any images/volumes into the model. |


???+ example "A typical AUCMEDI pipeline"
    ```python
    # AUCMEDI library
    from aucmedi import *

    # Pillar #1: Initialize input data reader
    ds = input_interface(interface="csv",
                         path_imagedir="dataset/images/",
                         path_data="dataset/classes.csv",
                         ohe=False, col_sample="ID", col_class="diagnosis")
    (index_list, class_ohe, nclasses, class_names, image_format) = ds

    # Pillar #2: Initialize a DenseNet121 model with ImageNet weights
    model = NeuralNetwork(n_labels=nclasses, channels=3,
                           architecture="2D.DenseNet121",
                           pretrained_weights=True)

    # Pillar #3: Initialize training Data Generator for first 1000 samples
    train_gen = DataGenerator(samples=index_list[:1000],
                              path_imagedir="dataset/images/",
                              labels=class_ohe[:1000],
                              image_format=image_format,
                              resize=model.meta_input,
                              standardize_mode=model.meta_standardize)
    # Run model training with Transfer Learning
    model.train(train_gen, epochs=20, transfer_learning=True)

    # Pillar #3: Initialize testing Data Generator for 500 samples
    test_gen = DataGenerator(samples=index_list[1000:1500],
                             path_imagedir="dataset/images/",
                             labels=None,
                             image_format=image_format,
                             resize=model.meta_input,
                             standardize_mode=model.meta_standardize)
    # Run model inference for unknown samples
    preds = model.predict(test_gen)
    ```
"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
from aucmedi.data_processing.io_data import input_interface
from aucmedi.data_processing.data_generator import DataGenerator
from aucmedi.data_processing.augmentation import ImageAugmentation, \
                                                 VolumeAugmentation, \
                                                 BatchgeneratorsAugmentation
from aucmedi.neural_network.model import NeuralNetwork



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import numpy as np
import tempfile
from PIL import Image
import SimpleITK as sitk
import os
#Internal libraries
from aucmedi.data_processing.io_loader import *
from aucmedi import DataGenerator

#-----------------------------------------------------#
#                 Unittest: IO Loader                 #
#-----------------------------------------------------#
class IOloaderTEST(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        self.img_2d_gray = np.random.rand(16, 16, 1) * 255
        self.img_2d_rgb = np.random.rand(16, 16, 3) * 255
        self.img_3d_gray = np.random.rand(16, 16, 16, 1) * 255
        self.img_3d_rgb = np.random.rand(16, 16, 16, 3) * 255
        self.img_3d_rgb = self.img_3d_rgb.astype(int)
        self.img_3d_hu = np.float32(np.random.rand(16, 16, 16, 1) * 1500 - 500)

    #-------------------------------------------------#
    #                  Image Loader                   #
    #-------------------------------------------------#
    # Test for DataGenerator functionality
    def test_image_loader_DataGenerator(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        # Create dataset
        sample_list = []
        for i in range(0, 6):
           img_pillow = Image.fromarray(self.img_2d_rgb.astype(np.uint8))
           index = "image.sample_" + str(i) + ".png"
           path_sample = os.path.join(tmp_data.name, index)
           img_pillow.save(path_sample)
           sample_list.append(index)
        # Test DataGenerator
        data_gen = DataGenerator(sample_list, tmp_data.name, resize=None,
                                 grayscale=False, batch_size=2)
        for i in range(0, 3):
            batch = data_gen[i]
            self.assertTrue(np.array_equal(batch[0].shape, (2, 16, 16, 3)))

    # Test for grayscale images
    def test_image_loader_2Dgray(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 5):
            # Create image
            img_store = np.squeeze(self.img_2d_gray, axis=-1)
            img_pillow = Image.fromarray(img_store.astype(np.uint8))
            index = "image.sample_" + str(i) + ".png"
            path_sample = os.path.join(tmp_data.name, index)
            img_pillow.save(path_sample)
            # Load image via loader
            img = image_loader(index, tmp_data.name, image_format=None,
                               grayscale=True)
            self.assertTrue(np.array_equal(img.shape, self.img_2d_gray.shape))

    # Test for rgb images
    def test_image_loader_2Drgb(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 5):
            # Create image
            img_pillow = Image.fromarray(self.img_2d_rgb.astype(np.uint8))
            index = "image.sample_" + str(i) + ".png"
            path_sample = os.path.join(tmp_data.name, index)
            img_pillow.save(path_sample)
            # Load image via loader
            img = image_loader(index, tmp_data.name, image_format=None,
                               grayscale=False)
            self.assertTrue(np.array_equal(img.shape, self.img_2d_rgb.shape))

    #-------------------------------------------------#
    #                  NumPy Loader                   #
    #-------------------------------------------------#
    # Test for DataGenerator functionality
    def test_numpy_loader_DataGenerator(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        # Create dataset
        sample_list = []
        for i in range(0, 6):
           index = "3Dimage.sample_" + str(i) + ".npy"
           path_sample = os.path.join(tmp_data.name, index)
           np.save(path_sample, self.img_3d_gray)
           sample_list.append(index)
        # Test DataGenerator
        data_gen = DataGenerator(sample_list, tmp_data.name, loader=numpy_loader,
                                 resize=None, two_dim=False, standardize_mode=None,
                                 grayscale=True, batch_size=2)
        for i in range(0, 3):
            batch = data_gen[i]
            self.assertTrue(np.array_equal(batch[0].shape, (2, 16, 16, 16, 1)))

    # Test for grayscale 2D images
    def test_numpy_loader_2Dgray(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 5):
            # Create image
            index = "image.sample_" + str(i) + ".npy"
            path_sample = os.path.join(tmp_data.name, index)
            np.save(path_sample, self.img_2d_gray)
            # Load image via loader
            img = numpy_loader(index, tmp_data.name, image_format=None,
                               grayscale=True, two_dim=True)
            self.assertTrue(np.array_equal(img.shape, self.img_2d_gray.shape))

    # Test for grayscale 3D images
    def test_numpy_loader_3Dgray(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 5):
            # Create image
            index = "image.sample_" + str(i) + ".npy"
            path_sample = os.path.join(tmp_data.name, index)
            np.save(path_sample, self.img_3d_gray)
            # Load image via loader
            img = numpy_loader(index, tmp_data.name, image_format=None,
                               grayscale=True, two_dim=False)
            self.assertTrue(np.array_equal(img.shape, self.img_3d_gray.shape))

    # Test for rgb 2D images
    def test_numpy_loader_2Drgb(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 5):
            # Create image
            index = "image.sample_" + str(i) + ".npy"
            path_sample = os.path.join(tmp_data.name, index)
            np.save(path_sample, self.img_2d_rgb)
            # Load image via loader
            img = numpy_loader(index, tmp_data.name, image_format=None,
                               grayscale=False, two_dim=True)
            self.assertTrue(np.array_equal(img.shape, self.img_2d_rgb.shape))

    # Test for rgb 3D images
    def test_numpy_loader_3Drgb(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 5):
            # Create image
            index = "image.sample_" + str(i) + ".npy"
            path_sample = os.path.join(tmp_data.name, index)
            np.save(path_sample, self.img_3d_rgb)
            # Load image via loader
            img = numpy_loader(index, tmp_data.name, image_format=None,
                               grayscale=False, two_dim=False)
            self.assertTrue(np.array_equal(img.shape, self.img_3d_rgb.shape))

    #-------------------------------------------------#
    #                   sITK Loader                   #
    #-------------------------------------------------#
    # Test for DataGenerator functionality
    def test_sitk_loader_DataGenerator(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        # Create mha dataset
        sample_list = []
        for i in range(0, 3):
            index = "3Dimage.sample_" + str(i) + ".mha"
            path_sample = os.path.join(tmp_data.name, index)
            image_sitk = sitk.GetImageFromArray(self.img_3d_hu)
            image_sitk.SetSpacing([0.5,1.5,2.0])
            sitk.WriteImage(image_sitk, path_sample)
            sample_list.append(index)
        # Create nii dataset
        for i in range(0, 3):
            index = "3Dimage.sample_" + str(i) + ".nii"
            path_sample = os.path.join(tmp_data.name, index)
            image_sitk = sitk.GetImageFromArray(self.img_3d_hu)
            image_sitk.SetSpacing([1.75,1.25,0.75])
            sitk.WriteImage(image_sitk, path_sample)
            sample_list.append(index)

        # Test DataGenerator
        data_gen = DataGenerator(sample_list, tmp_data.name,
                                 loader=sitk_loader,
                                 resize=None, standardize_mode=None,
                                 grayscale=True, batch_size=1)
        for i in range(0, 6):
            batch = data_gen[i]
            if i < 3:
                self.assertTrue(np.array_equal(batch[0].shape, (1, 32, 24, 8, 1)))
            else:
                self.assertTrue(np.array_equal(batch[0].shape, (1, 12, 20, 28, 1)))

    # Test for rgb 3D images
    def test_sitk_loader_3Drgb(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        # Run analysis
        for i in range(0, 6):
            if i < 3: format = ".mha"
            else : format = ".nii"
            # Create image
            index = "3Dimage.sample_" + str(i) + format
            path_sample = os.path.join(tmp_data.name, index)
            image_sitk = sitk.GetImageFromArray(self.img_3d_rgb)
            image_sitk.SetSpacing([0.5,1.5,2.0])
            sitk.WriteImage(image_sitk, path_sample)
            # Load image via loader
            img = sitk_loader(index, tmp_data.name, image_format=None)
            self.assertTrue(np.array_equal(img.shape, (32, 24, 8, 3)))

    # Test for hu 3D images
    def test_sitk_loader_3Dhu(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        # Run analysis
        for i in range(0, 6):
            if i < 3: format = ".mha"
            else : format = ".nii"
            # Create image
            index = "3Dimage.sample_" + str(i) + format
            path_sample = os.path.join(tmp_data.name, index)
            image_sitk = sitk.GetImageFromArray(self.img_3d_hu)
            image_sitk.SetSpacing([0.5,1.5,2.0])
            sitk.WriteImage(image_sitk, path_sample)
            # Load image via loader
            img = sitk_loader(index, tmp_data.name, image_format=None)
            self.assertTrue(np.array_equal(img.shape, (32, 24, 8, 1)))

    # Test for Resampling
    def test_sitk_loader_Resampling(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        # Run analysis
        sample_list = []
        for i in range(0, 4):
            if i < 2: format = ".mha"
            else : format = ".nii"
            # Create image
            index = "3Dimage.sample_" + str(i) + format
            path_sample = os.path.join(tmp_data.name, index)
            image_sitk = sitk.GetImageFromArray(self.img_3d_hu)
            image_sitk.SetSpacing([0.5,0.5,2.0])
            sitk.WriteImage(image_sitk, path_sample)
            sample_list.append(index)


        # Load images with Resampling 1x1x1
        for index in sample_list:
            img = sitk_loader(index, tmp_data.name, image_format=None)
            self.assertTrue(np.array_equal(img.shape, (32, 8, 8, 1)))
        # Load images with Resampling 0.5x0.5x2.0
        for index in sample_list:
            img = sitk_loader(index, tmp_data.name, image_format=None,
                              resampling=(2.0,0.5,0.5))
            self.assertTrue(np.array_equal(img.shape, (16, 16, 16, 1)))
        # Load images with Resampling None
        for index in sample_list:
            img = sitk_loader(index, tmp_data.name, image_format=None,
                              resampling=None)
            self.assertTrue(np.array_equal(img.shape, (16, 16, 16, 1)))
        # Load images via DataGenerator
        data_gen = DataGenerator(sample_list, tmp_data.name,
                                 loader=sitk_loader, resampling=(1.75,0.75,0.75),
                                 resize=None, standardize_mode=None,
                                 grayscale=True, batch_size=1)
        for i in range(0, 6):
            batch = data_gen[i]
            self.assertTrue(np.array_equal(batch[0].shape, (1, 18, 10, 10, 1)))

    #-------------------------------------------------#
    #                  Cache Loader                   #
    #-------------------------------------------------#
    # Test for DataGenerator functionality
    def test_cache_loader_DataGenerator(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        # Create dataset
        sample_list = []
        cache = {}
        for i in range(0, 6):
           index = "3Dimage.sample_" + str(i)
           cache[index] = self.img_3d_gray
           sample_list.append(index)
        # Test DataGenerator
        data_gen = DataGenerator(sample_list, tmp_data.name, loader=cache_loader,
                                 resize=None, two_dim=False, standardize_mode=None,
                                 grayscale=True, batch_size=2, cache=cache)
        for i in range(0, 3):
            batch = data_gen[i]
            self.assertTrue(np.array_equal(batch[0].shape, (2, 16, 16, 16, 1)))

    # Test for grayscale 2D images
    def test_cache_loader_2Dgray(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 5):
            # Create image
            index = "image.sample_" + str(i)
            cache = {index: self.img_2d_gray}
            # Load image via loader
            img = cache_loader(index, tmp_data.name, image_format=None,
                               grayscale=True, two_dim=True, cache=cache)
            self.assertTrue(np.array_equal(img.shape, self.img_2d_gray.shape))

    # Test for grayscale 3D images
    def test_cache_loader_3Dgray(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 5):
            # Create image
            index = "image.sample_" + str(i)
            cache = {index: self.img_3d_gray}
            # Load image via loader
            img = cache_loader(index, tmp_data.name, image_format=None,
                               grayscale=True, two_dim=False, cache=cache)
            self.assertTrue(np.array_equal(img.shape, self.img_3d_gray.shape))

    # Test for rgb 2D images
    def test_cache_loader_2Drgb(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 5):
            # Create image
            index = "image.sample_" + str(i)
            cache = {index: self.img_2d_rgb}
            # Load image via loader
            img = cache_loader(index, tmp_data.name, image_format=None,
                               grayscale=False, two_dim=True, cache=cache)
            self.assertTrue(np.array_equal(img.shape, self.img_2d_rgb.shape))

    # Test for rgb 3D images
    def test_cache_loader_3Drgb(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 5):
            # Create image
            index = "image.sample_" + str(i)
            cache = {index: self.img_3d_rgb}
            # Load image via loader
            img = cache_loader(index, tmp_data.name, image_format=None,
                               grayscale=False, two_dim=False, cache=cache)
            self.assertTrue(np.array_equal(img.shape, self.img_3d_rgb.shape))

    # Test for hu 3D images
    def test_cache_loader_3Dhu(self):
        # Create temporary directory
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 5):
            # Create image
            index = "image.sample_" + str(i)
            cache = {index: self.img_3d_hu}
            # Load image via loader
            img = cache_loader(index, tmp_data.name, image_format=None,
                               grayscale=True, two_dim=False, cache=cache)
            self.assertTrue(np.array_equal(img.shape, self.img_3d_hu.shape))

    # Test for Exception
    def test_cache_loader_exception(self):
        self.assertRaises(TypeError, cache_loader, index="test")



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import albumentations
import volumentations
# Internal libraries/scripts
from aucmedi.data_processing.subfunctions.sf_base import Subfunction_Base

#-----------------------------------------------------#
#             Subfunction class: Cropping             #
#-----------------------------------------------------#
class Crop(Subfunction_Base):
    """ A Crop Subfunction class which center/randomly crops a desired shape from an image.

    List of valid modes for parameter "mode": `["center", "random"]`

    ???+ info "2D image"
        Shape have to be defined as tuple with x and y size: `Crop(shape=(224, 224))`

        Cropping is done via albumentations CenterCrop and RandomCrop transform. <br>
        https://albumentations.ai/docs/api_reference/augmentations/crops/transforms/#albumentations.augmentations.crops.transforms.CenterCrop  <br>
        https://albumentations.ai/docs/api_reference/augmentations/crops/transforms/#albumentations.augmentations.crops.transforms.RandomCrop  <br>

    ???+ info "3D volume"
        Shape has to be defined as tuple with x, y and z size: `Crop(shape=(224, 224, 244))`

        Cropping is done via volumentations CenterCrop and RandomCrop transform. <br>
        https://github.com/muellerdo/volumentations
    """
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    def __init__(self, shape=(224, 224), mode="center"):
        """ Initialization function for creating a Crop Subfunction which can be passed to a
            [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].

        Args:
            shape (tuple of int):       Desired output shape after cropping.
            mode (str):                 Selected mode for cropping.
        """
        # Initialize parameter
        params = {"p":1.0, "always_apply":True}
        # Select augmentation module and add further parameter depending on dimension
        if len(shape) == 2:
            params["height"] = shape[0]
            params["width"] = shape[1]
            mod = albumentations
        elif len(shape) == 3:
            params["shape"] = shape
            mod = volumentations
        else : raise ValueError("Shape for cropping has to be 2D or 3D!", shape)
        # Initialize cropping transform
        if mode == "center":
            self.aug_transform = mod.Compose([mod.CenterCrop(**params)])
        elif mode == "random":
            self.aug_transform = mod.Compose([mod.RandomCrop(**params)])
        else : raise ValueError("Unknown mode for crop Subfunction", mode,
                                "Possibles modes are: ['center', 'random']")
        # Cache shape
        self.shape = shape

    #---------------------------------------------#
    #                Transformation               #
    #---------------------------------------------#
    def transform(self, image):
        # Perform cropping on image
        image_cropped = self.aug_transform(image=image)["image"]
        # Return cropped image
        return image_cropped



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
import tensorflow as tf
# Internal Libraries
from aucmedi.xai.methods.xai_base import XAImethod_Base

#-----------------------------------------------------#
#                Occlusion Sensitivity                #
#-----------------------------------------------------#
class OcclusionSensitivity(XAImethod_Base):
    """ XAI Method for Occlusion Sensitivity.

    Normally, this class is used internally in the [aucmedi.xai.decoder.xai_decoder][] in the AUCMEDI XAI module.

    ??? abstract "Reference - Implementation"
        Author: Raphael Meudec <br>
        GitHub Profile: [https://gist.github.com/RaphaelMeudec](https://gist.github.com/RaphaelMeudec) <br>
        Date: Jul 18, 2019 <br>
        [https://gist.github.com/RaphaelMeudec/7985b0c5eb720a29021d52b0a0be549a](https://gist.github.com/RaphaelMeudec/7985b0c5eb720a29021d52b0a0be549a) <br>

    This class provides functionality for running the compute_heatmap function,
    which computes a Occlusion Sensitivity Map for an image with a model.
    """
    def __init__(self, model, layerName=None, patch_size=16):
        """ Initialization function for creating a Occlusion Sensitivity Map as XAI Method object.

        Args:
            model (keras.model):            Keras model object.
            layerName (str):                Not required in Occlusion Sensitivity Maps, but defined by Abstract Base Class.
        """
        # Cache class parameters
        self.model = model
        self.patch_size = patch_size

    #---------------------------------------------#
    #             Heatmap Computation             #
    #---------------------------------------------#
    def compute_heatmap(self, image, class_index, eps=1e-8):
        """ Core function for computing the Occlusion Sensitivity Map for a provided image and for specific classification outcome.

        ???+ attention
            Be aware that the image has to be provided in batch format.

        Args:
            image (numpy.ndarray):              Image matrix encoded as NumPy Array (provided as one-element batch).
            class_index (int):                  Classification index for which the heatmap should be computed.
            eps (float):                        Epsilon for rounding.

        The returned heatmap is encoded within a range of [0,1]

        ???+ attention
            The shape of the returned heatmap is 2D -> batch and channel axis will be removed.

        Returns:
            heatmap (numpy.ndarray):            Computed Occlusion Sensitivity Map for provided image.
        """
        # Utilize only image matrix instead of batch
        image = image[0]
        # Create empty sensitivity map
        sensitivity_map = np.zeros((image.shape[0], image.shape[1]))
        # Iterate the patch over the image
        for top_left_x in range(0, image.shape[0], self.patch_size):
            for top_left_y in range(0, image.shape[1], self.patch_size):
                patch = apply_grey_patch(image, top_left_x, top_left_y,
                                         self.patch_size)
                prediction = self.model.predict(np.array([patch]))[0]
                confidence = prediction[class_index]

                # Save confidence for this specific patch in the map
                sensitivity_map[
                    top_left_y:top_left_y + self.patch_size,
                    top_left_x:top_left_x + self.patch_size,
                ] = 1 - confidence
        # Return the resulting sensitivity map (automatically a heatmap)
        return sensitivity_map

#-----------------------------------------------------#
#                     Subroutines                     #
#-----------------------------------------------------#
def apply_grey_patch(image, top_left_x, top_left_y, patch_size):
    """ Internal function.

    Replace a part of the image with a grey patch.

    Args:
        image (numpy.ndarray):                  Input image
        top_left_x (int):                       Top Left X position of the applied box
        top_left_y (int):                       Top Left Y position of the applied box
        patch_size (int):                       Size of patch to apply

    Returns:
        patched_image (numpy.ndarray):          Patched image
    """
    patched_image = np.array(image, copy=True)
    patched_image[top_left_y:top_left_y + patch_size, top_left_x:top_left_x + patch_size, :] = 127.5
    return patched_image



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
import os
# AUCMEDI Libraries
from aucmedi.xai.methods import xai_dict
from aucmedi.data_processing.io_loader import image_loader
from aucmedi.utils.visualizer import visualize_heatmap
from aucmedi.data_processing.subfunctions import Resize

#-----------------------------------------------------#
#                    XAI - Decoder                    #
#-----------------------------------------------------#
def xai_decoder(data_gen, model, preds=None, method="gradcam", layerName=None,
                alpha=0.4, out_path=None):
    """ XAI Decoder function for automatic computation of Explainable AI heatmaps.

    This module allows to visualize which regions were crucial for the neural network model
    to compute a classification on the provided unknown images.

    - If `out_path` parameter is None, heatmaps are returned as NumPy array.
    - If a path is provided as `out_path`, then heatmaps are stored to disk as PNG files.

    ???+ info "XAI Methods"
        The XAI Decoder can be run with different XAI methods as backbone.

        A list of all implemented methods and their keys can be found here: <br>
        [aucmedi.xai.methods][]

    ???+ example "Example"
        ```python
        # Create a DataGenerator for data I/O
        datagen = DataGenerator(samples[:3], "images_xray/", labels=None, resize=(299, 299))

        # Get a model
        model = NeuralNetwork(n_labels=3, channels=3, architecture="Xception",
                               input_shape=(299,299))
        model.load("model.xray.keras")

        # Make some predictions
        preds = model.predict(datagen)

        # Compute XAI heatmaps via Grad-CAM (resulting heatmaps are stored in out_path)
        xai_decoder(datagen, model, preds, method="gradcam", out_path="xai.xray_gradcam")
        ```

    Args:
        data_gen (DataGenerator):           A data generator which will be used for inference.
        model (NeuralNetwork):             Instance of a AUCMEDI neural network class.
        preds (numpy.ndarray):              NumPy Array of classification prediction encoded as OHE (output of a AUCMEDI prediction).
        method (str):                       XAI method class instance or index. By default, GradCAM is used as XAI method.
        layerName (str):                    Layer name of the convolutional layer for heatmap computation. If `None`, the last conv layer is used.
        alpha (float):                      Transparency value for heatmap overlap plotting on input image (range: [0-1]).
        out_path (str):                     Output path in which heatmaps are saved to disk as PNG files.

    Returns:
        images (numpy.ndarray):             Combined array of images. Will be only returned if `out_path` parameter is `None`.
        heatmaps (numpy.ndarray):           Combined array of XAI heatmaps. Will be only returned if `out_path` parameter is `None`.
    """
    # Initialize & access some variables
    batch_size = data_gen.batch_size
    n_classes = model.n_labels
    sample_list = data_gen.samples
    # Prepare XAI output methods
    res_img = []
    res_xai = []
    if out_path is not None and not os.path.exists(out_path) : os.mkdir(out_path)
    # Initialize xai method
    if isinstance(method, str) and method in xai_dict:
        xai_method = xai_dict[method](model.model, layerName=layerName)
    else : xai_method = method

    # Iterate over all samples
    for i in range(0, len(sample_list)):
        # Load original image
        img_org = image_loader(sample_list[i], data_gen.path_imagedir,
                               image_format=data_gen.image_format,
                               grayscale=data_gen.grayscale)
        shape_org = img_org.shape[0:2]
        # Load processed image
        img_prc = data_gen.preprocess_image(i)
        img_batch = np.expand_dims(img_prc, axis=0)
        # If preds given, compute heatmap only for argmax class
        if preds is not None:
            ci = np.argmax(preds[i])
            xai_map = xai_method.compute_heatmap(img_batch, class_index=ci)
            xai_map = Resize(shape=shape_org).transform(xai_map)
            postprocess_output(sample_list[i], img_org, xai_map, n_classes,
                               data_gen, res_img, res_xai, out_path, alpha)
        # If no preds given, compute heatmap for all classes
        else:
            sample_maps = []
            for ci in range(0, n_classes):
                xai_map = xai_method.compute_heatmap(img_batch, class_index=ci)
                xai_map = Resize(shape=shape_org).transform(xai_map)
                sample_maps.append(xai_map)
            sample_maps = np.array(sample_maps)
            postprocess_output(sample_list[i], img_org, sample_maps, n_classes,
                               data_gen, res_img, res_xai, out_path, alpha)
    # Return output directly if no output path is defined (and convert to NumPy)
    if out_path is None : return np.array(res_img), np.array(res_xai)

#-----------------------------------------------------#
#          Subroutine: Output Postprocessing          #
#-----------------------------------------------------#
""" Helper/Subroutine function for XAI Decoder.

Caches heatmap for direct output or generates a visualization as PNG.
"""
def postprocess_output(sample, image, xai_map, n_classes, data_gen,
                       res_img, res_xai, out_path, alpha):
    # Update result lists for direct output
    if out_path is None:
        res_img.append(image)
        res_xai.append(xai_map)
    # Generate XAI heatmap visualization
    else:
        # Create XAI path
        if data_gen.image_format:
            xai_file = sample + "." + data_gen.image_format
        else : xai_file = sample
        if os.sep in xai_file : xai_file = xai_file.replace(os.sep, ".")
        path_xai = os.path.join(out_path, xai_file)
        # If preds given, output only argmax class heatmap
        if len(xai_map.shape) == 2:
            visualize_heatmap(image, xai_map, out_path=path_xai, alpha=alpha)
        # If no preds given, output heatmaps for all classes
        else:
            for c in range(0, n_classes):
                path_xai_c = path_xai[:-4] + ".class_" + str(c) + \
                             path_xai[-4:]
                visualize_heatmap(image, xai_map[c], out_path=path_xai_c,
                                  alpha=alpha)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import numpy as np
from sklearn.model_selection import StratifiedKFold, KFold
# Internal libraries
from aucmedi.sampling.iterative import MultilabelStratifiedKFold

#-----------------------------------------------------#
#    Function: Sampling via k-fold cross-validation   #
#-----------------------------------------------------#
def sampling_kfold(samples, labels, metadata=None, n_splits=3,
                   stratified=True, iterative=False, seed=None):
    """ Simple wrapper function for calling k-fold cross-validation sampling functions.

    Allow usage of stratified and iterative sampling algorithm.

    ???+ warning
        Be aware that multi-label data does not support random stratified sampling.

    ???+ example
        The sampling is returned as list with length n_splits containing tuples with sampled data.

        ```python title="Example for n_splits=3"
        cv = sampling_kfold(samples, labels, n_splits=3)

        # sampling in which x = samples and y = labels
        # cv <-> [(train_x, train_y, test_x, test_y),   # fold 1
        #         (train_x, train_y, test_x, test_y),   # fold 2
        #         (train_x, train_y, test_x, test_y)]   # fold 3

        # Recommended access on the folds
        for fold in cv:
            (train_x, train_y, test_x, test_y) = fold
        ```

        ```python title="Example with metadata"
        cv = sampling_kfold(samples, labels, metadata, n_splits=2)

        # sampling in which x = samples, y = labels and m = metadata
        # cv <-> [(train_x, train_y, train_m, test_x, test_y, test_m),      # fold 1
        #         (train_x, train_y, train_m, test_x, test_y, test_m)]      # fold 2
        ```

    Args:
        samples (list of str):      List of sample/index encoded as Strings.
        labels (numpy.ndarray):     NumPy matrix containing the ohe encoded classification.
        metadata (numpy.ndarray):   NumPy matrix with additional metadata. Have to be shape (n_samples, meta_variables).
        n_splits (int):             Number of folds (k). Must be at least 2.
        stratified (bool):          Option whether to use stratified sampling based on provided labels.
        iterative (bool):           Option whether to use iterative sampling algorithm.
        seed (int):                 Seed to ensure reproducibility for random functions.

    Returns:
        sampling (list of tuple):   List with length `n_splits` containing tuples with sampled data.
    """
    # Initialize variables
    results = []
    wk_labels = labels

    # Initialize random sampler
    if not stratified and not iterative:
        sampler = KFold(n_splits=n_splits, shuffle=True, random_state=seed)
    # Initialize random stratified sampler
    elif stratified and not iterative:
        sampler = StratifiedKFold(n_splits=n_splits, shuffle=True,
                                  random_state=seed)
        wk_labels = np.argmax(wk_labels, axis=-1)
    # Initialize iterative stratified sampler
    else:
        sampler = MultilabelStratifiedKFold(n_splits=n_splits, shuffle=True,
                                            random_state=seed)

    # Preprocess data
    x = np.asarray(samples)
    y = np.asarray(labels)
    if metadata is not None : m = np.asarray(metadata)

    # Apply sampling and generate folds
    for train, test in sampler.split(X=samples, y=wk_labels):
        # Simple sampling
        if metadata is None:
            fold = (x[train], y[train], x[test], y[test])
        # Sampling with metadata
        else:
            fold = (x[train], y[train], m[train], x[test], y[test], m[test])
        results.append(fold)

    # Return result sampling
    return results



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import numpy as np
import pandas as pd
import tempfile
from PIL import Image
import json
import os
#Internal libraries
from aucmedi.data_processing.io_interfaces import *

#-----------------------------------------------------#
#               Unittest: IO Interfaces               #
#-----------------------------------------------------#
class IOinterfacesTEST(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        self.aif = ["png"]

    #-------------------------------------------------#
    #              Directory IO Interface             #
    #-------------------------------------------------#
    def test_Directory_testing(self):
        # Create imaging data
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 25):
            img = np.random.rand(16, 16, 3) * 255
            img_pillow = Image.fromarray(img.astype(np.uint8))
            index = "image.sample_" + str(i) + ".png"
            path_sample = os.path.join(tmp_data.name, index)
            img_pillow.save(path_sample)
        # Run Directory IO
        ds = directory_loader(tmp_data.name, self.aif, training=False)
        self.assertTrue(len(ds[0]), 25)

    def test_Directory_training(self):
        # Create imaging data with subdirectories
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        for i in range(0, 5):
            os.mkdir(os.path.join(tmp_data.name, "class_" + str(i)))
        # Fill subdirectories with images
        for i in range(0, 25):
            img = np.random.rand(16, 16, 3) * 255
            img_pillow = Image.fromarray(img.astype(np.uint8))
            index = "image.sample_" + str(i) + ".png"
            label_dir = "class_" + str((i % 5))
            path_sample = os.path.join(tmp_data.name, label_dir, index)
            img_pillow.save(path_sample)
        # Run Directory IO
        ds = directory_loader(tmp_data.name, self.aif, training=True)
        self.assertTrue(len(ds[0]), 25)
        self.assertTrue(len(ds[1]), 25)

    #-------------------------------------------------#
    #                JSON IO Interface                #
    #-------------------------------------------------#
    def test_JSON_testing(self):
        # Create imaging data
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        data = {}
        for i in range(0, 25):
            img = np.random.rand(16, 16, 3) * 255
            img_pillow = Image.fromarray(img.astype(np.uint8))
            index = "image.sample_" + str(i) + ".png"
            data[index[:-4]] = 0
            path_sample = os.path.join(tmp_data.name, index)
            img_pillow.save(path_sample)
        # Create JSON data
        tmp_json = tempfile.NamedTemporaryFile(mode="w", prefix="tmp.aucmedi.",
                                               suffix=".json")
        json.dump(data, tmp_json)
        tmp_json.flush()
        # Run JSON IO
        ds = json_loader(path_data=tmp_json.name, path_imagedir=tmp_data.name,
                         allowed_image_formats=self.aif, training=False)
        self.assertTrue(len(ds[0]), 25)

    def test_JSON_training_woOHE(self):
        # Create imaging data with subdirectories
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        data = {}
        for i in range(0, 25):
            img = np.random.rand(16, 16, 3) * 255
            img_pillow = Image.fromarray(img.astype(np.uint8))
            index = "image.sample_" + str(i) + ".png"
            data[index[:-4]] = np.random.randint(5)
            path_sample = os.path.join(tmp_data.name, index)
            img_pillow.save(path_sample)
        # Create JSON data
        tmp_json = tempfile.NamedTemporaryFile(mode="w", prefix="tmp.aucmedi.",
                                               suffix=".json")
        json.dump(data, tmp_json)
        tmp_json.flush()
        # Run JSON IO
        ds = json_loader(path_data=tmp_json.name, path_imagedir=tmp_data.name,
                         allowed_image_formats=self.aif, training=True,
                         ohe=False)
        self.assertTrue(len(ds[0]), 25)
        self.assertTrue(len(ds[1]), 25)

    def test_JSON_training_withOHE(self):
        # Create imaging data with subdirectories
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        data = {}
        for i in range(0, 25):
            img = np.random.rand(16, 16, 3) * 255
            img_pillow = Image.fromarray(img.astype(np.uint8))
            index = "image.sample_" + str(i) + ".png"
            labels_ohe = [0, 0, 0, 0]
            class_index = np.random.randint(0, 4)
            labels_ohe[class_index] = 1
            data[index[:-4]] = labels_ohe
            path_sample = os.path.join(tmp_data.name, index)
            img_pillow.save(path_sample)
        # Create JSON data
        tmp_json = tempfile.NamedTemporaryFile(mode="w", prefix="tmp.aucmedi.",
                                               suffix=".json")
        json.dump(data, tmp_json)
        tmp_json.flush()
        # Run JSON IO
        ds = json_loader(path_data=tmp_json.name, path_imagedir=tmp_data.name,
                         allowed_image_formats=self.aif, training=True,
                         ohe=True)
        self.assertTrue(len(ds[0]), 25)
        self.assertTrue(len(ds[1]), 25)

    #-------------------------------------------------#
    #                 CSV IO Interface                #
    #-------------------------------------------------#
    def test_CSV_testing(self):
        # Create imaging data
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        data = {}
        for i in range(0, 25):
            img = np.random.rand(16, 16, 3) * 255
            img_pillow = Image.fromarray(img.astype(np.uint8))
            index = "image.sample_" + str(i) + ".png"
            data[index[:-4]] = 0
            path_sample = os.path.join(tmp_data.name, index)
            img_pillow.save(path_sample)
        # Create CSV data
        tmp_csv = tempfile.NamedTemporaryFile(mode="w", prefix="tmp.aucmedi.",
                                              suffix=".csv")
        df = pd.DataFrame.from_dict(data, orient="index", columns=["class"])
        df.index.name = "index"
        df.to_csv(tmp_csv.name, index=True, header=True)

        # Run CSV IO
        ds = csv_loader(path_data=tmp_csv.name, path_imagedir=tmp_data.name,
                        allowed_image_formats=self.aif, training=False,
                        col_sample="index")
        self.assertTrue(len(ds[0]), 25)

    def test_CSV_training_woOHE(self):
        # Create imaging data with subdirectories
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        data = {}
        for i in range(0, 25):
            img = np.random.rand(16, 16, 3) * 255
            img_pillow = Image.fromarray(img.astype(np.uint8))
            index = "image.sample_" + str(i) + ".png"
            data[index[:-4]] = np.random.randint(5)
            path_sample = os.path.join(tmp_data.name, index)
            img_pillow.save(path_sample)
        # Create CSV data
        tmp_csv = tempfile.NamedTemporaryFile(mode="w", prefix="tmp.aucmedi.",
                                              suffix=".csv")
        df = pd.DataFrame.from_dict(data, orient="index", columns=["class"])
        df.index.name = "index"
        df.to_csv(tmp_csv.name, index=True, header=True)

        # Run CSV IO
        ds = csv_loader(path_data=tmp_csv.name, path_imagedir=tmp_data.name,
                        allowed_image_formats=self.aif, training=True,
                        ohe=False, col_sample="index", col_class="class")
        self.assertTrue(len(ds[0]), 25)
        self.assertTrue(len(ds[1]), 25)

    def test_CSV_training_withOHE(self):
        # Create imaging data with subdirectories
        tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                               suffix=".data")
        data = {}
        for i in range(0, 25):
            img = np.random.rand(16, 16, 3) * 255
            img_pillow = Image.fromarray(img.astype(np.uint8))
            index = "image.sample_" + str(i) + ".png"
            labels_ohe = [0, 0, 0, 0]
            class_index = np.random.randint(0, 4)
            labels_ohe[class_index] = 1
            data[index[:-4]] = labels_ohe
            path_sample = os.path.join(tmp_data.name, index)
            img_pillow.save(path_sample)
        # Create CSV data
        tmp_csv = tempfile.NamedTemporaryFile(mode="w", prefix="tmp.aucmedi.",
                                              suffix=".csv")
        df = pd.DataFrame.from_dict(data, orient="index",
                                    columns=["a", "b", "c", "d"])
        df.index.name = "index"
        df.to_csv(tmp_csv.name, index=True, header=True)

        # Run CSV IO
        ds = csv_loader(path_data=tmp_csv.name, path_imagedir=tmp_data.name,
                        allowed_image_formats=self.aif, training=True,
                        ohe=True, col_sample="index")
        self.assertTrue(len(ds[0]), 25)
        self.assertTrue(len(ds[1]), 25)
